`gpcm` <-
function(X,prior.dist='norm',tol1=1e-4,tol2=0.5*tol1,
  n.quad.pts=41,...){
  # X is an N x J matrix of item responses
  # prior.dist is the prior density function for theta
  # ..., are parameters passed to the prior density
  # Choices are determined by the guass.quad.prob function
  # in library(statmod)---uniform, norm, beta, gamma
  # for the time being the parameters of the prior dist are
  # fixed constants (not estimated).
  
  # Constants 
  K = apply(X,2,max,na.rm=T)
  J = ncol(X)
  N = nrow(X)
  df = sum(K)+J
  ###################################################################
  # Initialize parameter values #
  ###################################################################
  a = rep(.1,J)
  b = rep(0,J)
  d = matrix(0,J,max(K)) 

  for(i in 1:J){
    if(K[i]<max(K))
      is.na(d[i,(K[i]+1):max(K)]) = T
  }
  # the matrix d is J x max(K), where NAs are used if a given item
  # has fewer than max(K)+1 score categories, maybe inefficient use
  # of memory, but simplifies calculations later

  ###################################################################


  ###################################################################
  # Find unique response patterns and count
  # This part cannot handle NAs
  ####################################################################
  U = as.matrix(unique(as.data.frame(X)))
  n = table(as.data.frame(X))
  n = n[1+U%*%c(1,cumprod(K[-J]+1))]
  ####################################################################
  
  GausHerm = gauss.quad.prob(n=9,dist=prior.dist)
  th = GausHerm$node
  wt = GausHerm$w

  ab = cbind(a,b,d)
  flag = FALSE
  while(T){
    while(T){
      # This part uses the gradient-EM approach#
      old = ab
      ab = update.gpcm.em(th,U,a,b,d,wt,n)
      a = ab[,1]
      b = ab[,2]
      d = ab[,-(1:2)]
      if(max(abs(ab-old),na.rm=T)<tol1) break;
    }
    if(flag ==TRUE) break;
    GausHerm = gauss.quad.prob(n=n.quad.pts,dist=prior.dist,...)
    th = GausHerm$node
    wt = GausHerm$w
    tol1 = tol2
    flag = TRUE
  }
  while(T){
    # This part uses Newton-Raphson (without EM) to update #
    # It also gets an estimate of the standard errors of the ests #
    old = ab
    out = update.gpcm.nr(th,U,a,b,d,wt,n)
    ab = out$est
    a = ab[,1]
    b = ab[,2]
    d = ab[,-(1:2)]
    if(max(abs(ab-old),na.rm=T)<tol2) break;
  }
  d = as.data.frame(d)
  names(d)=paste('Cat',1:max(K))
  parms = data.frame(Slope=a,Intercept=b,d)
  row.names(parms)=paste('Item',1:J)
  log.like = sum(n*log(out$p))
  eap = out$p.x%*%th/out$p
  out = list(estimates = parms, se = out$se, data.unique = U,
    cts = n, prior.dist = prior.dist,
    call = match.call(),log.like=log.like,df=df,eap=eap)
  out$se = as.data.frame(out$se)
  row.names(out$se) = row.names(parms)
  names(out$se) = paste('SE.',c('a','b',paste('d',1:max(K),sep='')),sep='')
  class(out) = c('gpcm','data.frame')
  out
}

